"""Satellite encounter windows and scheduling.

Notes
-----
This module contains the Encounter class to determine and schedule
satellite-earth encounters.

Class's
-------
Encounter: Object used to calculate and store encounter information.
    addEncounter : Defines and stores EncounterSpec object in encounters
        attribute.
    getSunPos : Instantiates sunPos attribute.
    getWindows : Instantiates windows attribute of EncounterSpec object.
    saveWindows : Saves window data in local directory.

EncounterSpec : Object used to localize GoundPosition encounter information.
"""


import numpy as np
import pandas as pd
import pkg_resources
import julian
from datetime import datetime
from jplephem.spk import SPK
from celest import Satellite



class Encounter:
    """
    Encounter object computes, schedules, and stores satellite encounters.

    Methods
    -------
    addEncounter : Defines and stores EncounterSpec object in encounters
        attribute.
    getSunPos : Instantiates sunPos attribute.
    getWindows : Instantiates windows attribute of EncounterSpec object.
    saveWindows : Saves window data in local directory.

    Instance Variables
    ------------------
    encounters : Dictionary of EncounterSpec objects.
    sunPos : Sun positions in ECEF frame.
    """

    def __init__(self):
        """Define instance variables."""

        self.encounters = {}
        self.sunPos = None

    def addEncounter(self, name, encType, groundPos, ang, angType, maxAng, solar=0):
        """
        Define an encounter type.

        Uses the input data to create a key/value pair in the dictionary of the
        encounters attribute.

        Parameters
        ----------
        name : str
            String acting as the encounter identifier.
        encType :  str
            Specifies encounter as either imaging, "IMG", or data linking,
            "DL".
        goundPos : GroundPosition object
            Ground location associated with the encounter.
        ang : int or float
            Angluar constraint for the encounter.
        angType : str
            String specifying the constraint angle as either the altitude,
            "alt", or nadir-LOS, "nadirLOS", angle type.
        maxAng : bool
            Defines the contraint angle as a maximum constraint if True or as
            minimum constraint if False. Note that the nadirLOS angle is
            measured to increase away from nadir.
        solar : int
            Defines sunlight constraint. -1 -> windows at night, 0 -> windows
            at day or night, 1 -> windows at day.

        Returns
        -------
        None
        """
        if angType == "nadirLOS":
            angType = 1
        elif angType == "alt":
            angType = 0

        encounter = EncounterSpec(name, encType, groundPos, ang, angType, maxAng, solar)
        self.encounters[name] = encounter

    def getSunPos(self, timeData):
        """
        Instantiate sunPos attribute.

        Uses de421 ephemeris to calculate the sun"s position in the ECEF frame
        at each time in timeData.

        Parameters
        ----------
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        self.sunPos : list
            Array of shape (n,3) with columns of X, Y, Z ECEF sun position
            data.
        """
        # Get sun position in ECI.
        julianTimes = np.zeros((timeData.size,))

        for i, time in enumerate(timeData):
            julianTimes[i] = julian.to_jd(datetime.strptime(time[:18],
                                          "%Y-%m-%d %H:%M:%S"))

        ephem = pkg_resources.resource_filename(__name__, 'data/de421.bsp')
        kernal = SPK.open(ephem)

        ssb2sun = kernal[0, 10].compute(julianTimes)
        ssb2eb = kernal[0, 3].compute(julianTimes)
        eb2e = kernal[3, 399].compute(julianTimes)
        e2sun = (ssb2sun - ssb2eb - eb2e).T

        # Convert sun position data to ECEF
        sun = Satellite()
        e2sunECEF = sun.getECEF(posData=e2sun, timeData=timeData)
        self.sunPos = e2sunECEF
        return self.sunPos

    def getWindows(self, satellite):
        """
        Instantiates windows attribute of EncounterSpec objects.

        Determines the windows for all encounter types specified in the
        encounters attribute.

        Parameters
        ----------
        satellite : Satellite object
            Satellite object with appropriate dependencies instantiated.

        Returns
        -------
        self.windows : dict
            Dictionary of key/value pairs where the keys are the encounter
            names and the values are the windows associated with a specific
            encounter type.

        Usage
        -----
        Satellite.times attribute must be instantiated. nadirAng must
        be instantiated for the ground position located in satellite.gs if the
        corresponding encounter is a "nadirLOS" type. alt and az must be
        instantiated for the ground position located in satellite.gs if the
        encounter is an "alt" type.
        """

        for i in self.encounters:

            encounter = self.encounters[i]
            groundPos = encounter.groundPos.name
            ang = encounter.ang
            angType = encounter.angType
            maxAng = encounter.maxAng
            solar = encounter.solar

            times = satellite.times

            alt = satellite.gs[groundPos].alt
            nadir = satellite.gs[groundPos].nadirAng

            if not angType and maxAng:
                winInd = np.where((0 < alt) & (alt < ang))[0]
            elif not angType and not maxAng:
                winInd = np.where(alt > ang)[0]
            elif angType and maxAng:
                winInd = np.where((nadir < ang) & (alt >= 0))[0]
            elif angType and not maxAng:
                winInd = np.where((nadir > ang) & (alt >= 0))[0]

            # Get sun position vector.
            if solar != 0:
                sunECEFpos = self.getSunPos(timeData=times)
                gndECEFpos = np.full((sunECEFpos.shape[0], 3),
                                     satellite.gs[groundPos].ECEFpos)
                dividend = np.einsum("ij, ij->i", sunECEFpos, gndECEFpos)
                divisor = np.multiply(np.linalg.norm(sunECEFpos, axis=1),
                                      np.linalg.norm(gndECEFpos, axis=1))
                arg = np.divide(dividend, divisor)
                ang = np.degrees(np.arccos(arg))

            # Find intersection of night indices and window indices.
            if solar == -1:
                nightInd = np.where(ang >= 90)
                winInd = np.intersect1d(winInd, nightInd)

            # Find intersection of day indices and window indices.
            if solar == 1:
                dayInd = np.where(ang < 90)
                winInd = np.intersect1d(winInd, dayInd)

            winIndArr = np.split(winInd, np.where(np.diff(winInd) != 1)[0]+1)
            windowTimes = np.empty((len(winIndArr), 3), dtype="U25")

            for j in range(len(winIndArr)):

                start = datetime.strptime(times[winIndArr[j][0]][:18],
                                          "%Y-%m-%d %H:%M:%S")
                end = datetime.strptime(times[winIndArr[j][-1]][:18],
                                        "%Y-%m-%d %H:%M:%S")

                windowTimes[j, 0] = start
                windowTimes[j, 1] = end
                windowTimes[j, 2] = (end-start).total_seconds()

            self.encounters[i].windows = windowTimes
            self.encounters[i].length = windowTimes.shape[0]

    def saveWindows(self, fileName, delimiter):
        """
        Save window data to local directory.

        Parameters
        ----------
        fileName : str
            File name of output file. Can be a .txt or .csv file.
        delimiter : str
            String of length 1 representing the feild delimiter for the output
            file.

        Returns
        -------
        None

        Usage
        -----
        It is recommended to use a tab delimiter for .txt files and comma
        delimiters for .csv files. Will return an error if fileName already
        exists in the current working directory.
        """
        encounterTypes = np.array([])
        encounterStart = np.array([])
        encounterEnd = np.array([])
        elapsedSec = np.array([])

        for key in self.encounters:
            encounterTypes = np.append(encounterTypes,
                                       np.full((self.encounters[key].length,),
                                               key))
            encounterStart = np.append(encounterStart,
                                       self.encounters[key].windows[:, 0])
            encounterEnd = np.append(encounterEnd,
                                     self.encounters[key].windows[:, 1])
            elapsedSec = np.append(elapsedSec,
                                   self.encounters[key].windows[:, 2])

        data = {}
        data["Encounter Start"] = pd.Series(encounterStart)
        data["Encounter End"] = pd.Series(encounterEnd)
        data["Elapsed Seconds"] = pd.Series(elapsedSec)
        data["Name"] = pd.Series(encounterTypes)

        df = pd.DataFrame(data)
        df.to_csv(fileName, sep=delimiter)

    def getStats(self):
        """
        Get encounter statistics.

        This method produces various statistics for each encounter and
        encounter type. The generated statistics include the raw number of
        viable passes, cumulative time, daily average counts, and the daily
        average time for each encounter and encounter type.

        Parameters
        ----------
        None

        Returns
        -------
        data : Pandas DataFrame
            Pandas DataFrame containing the statistics for each encounter and
            encounter type.

        Usage
        -----
        The returned pandas DataFrame can be printed for easy viewing of the
        statistics.
        """
        data = {}

        numDL = 0
        timeDL = 0
        avgNumDL = 0
        avgTimeDL = 0

        numIMG = 0
        timeIMG = 0
        avgNumIMG = 0
        avgTimeIMG = 0

        for encounter in self.encounters:

            encType = self.encounters[encounter].type
            windows = self.encounters[encounter].windows
            length = round(self.encounters[encounter].length, 2)
            time = round(np.sum(windows[:, 2].astype(float)), 2)

            start = datetime.strptime(windows[0, 0], "%Y-%m-%d %H:%M:%S")
            end = datetime.strptime(windows[-1, 1], "%Y-%m-%d %H:%M:%S")
            deltaT = (end-start).days

            avgNum = round(length / deltaT, 2)
            avgTime = round(time / length, 2)

            if encType == "DL":
                numDL += length
                timeDL += time
                avgNumDL += avgNum
                avgTimeDL += avgTime

            if encType == "IMG":
                numIMG += length
                timeIMG += time
                avgNumIMG += avgNum
                avgTimeIMG += avgTime

            data[encounter] = pd.Series([length, avgNum, time, avgTime])

        data["Total DL"] = pd.Series([numDL, avgNumDL, timeDL, avgTimeDL])
        data["Total IMG"] = pd.Series([numIMG, avgNumIMG, timeIMG, avgTimeIMG])
        df = pd.DataFrame.from_dict(data)
        df.index = ["Number of Viable Encounters",
                    "Average Encounters per Day",
                    "Viable Encounters Duration (s)",
                    "Average Encounter Duration (s)"]

        return df


class EncounterSpec:
    """
    EncounterSpec object localizes encounter information specific to a
    GroundPosition Object. Units are in degrees and meters.

    Methods
    -------
    None

    Instance Variables
    ------------------
    name : Encounter identifier.
    type : Encounter type.
    groundPos : GroundPosition object related to the encounter.
    ang : Constraint angle for a viable encounter.
    angType : Constraint angle type.
    maxAng : Direction of the angular constraint.
    solar : Encounter daylight requirement.
    windows : Encounter Windows.
    length : Length of data attributes.
    """

    def __init__(self, name, encType, groundPos, ang, angType, maxAng, solar=0):
        """
        Define instance variables.

        Parameters
        ----------
        name : str
            String acting as the encounter identifier.
        encType :  str
            Specifies encounter as either imaging, "IMG", or data linking,
            "DL".
        goundPos : GroundPosition object
            Ground location associated with the encounter.
        ang : int or float
            Angluar constraint for the encounter.
        angType : str
            String specifying the constraint angle as either the altitude,
            "alt", or nadir-LOS, "nadirLOS", angle type.
        maxAng : bool
            Defines the contraint angle as a maximum constraint if True or as
            minimum constraint if False. Note that the nadirLOS angle is
            measured to increase away from nadir.
        solar : int
            Defines sunlight constraint. -1 -> windows at night, 0 -> windows
            at day or night, 1 -> windows at day.
        """

        self.name = name
        self.type = encType
        self.groundPos = groundPos
        self.ang = ang
        self.angType = angType
        self.maxAng = maxAng
        self.solar = solar
        self.windows = None
        self.length = 0

    def __str__(self):
        """Defines EncounterSpec informaiton string."""
        data = np.array([self.name, self.type, self.groundPos.name, self.ang,
                         self.angType, self.maxAng, self.solar])
        index = np.array(["Name:", "Encounter Type:", "Ground Position Name:",
                          "Constraint Angle:", "Angle Type:", "Is Max:",
                          "Solar Constraint:"])

        df = pd.DataFrame(data=data, index=index)
        return df.to_string()



"""Ground Position Instantiation.

Notes
-----
This module contains the GroundPosition class to instantiate ground position
data for use in the Satellite class.

Class
-----
GroundPosition: Object used to combine ground position data.
    getRadius : Used to instantiate the radius attribute.
"""


from math import sin, cos, sqrt, radians
import numpy as np


class GroundPosition:
    """
    GoundPosition object stores ground location data. Used in the Satellite
    class for altitude, azimuth, and nadir angle calculations.

    Methods
    -------
    getRadius : Used to instantiate the radius attribute.

    Instance Variables
    ------------------
    name : Tag for ground station location.
    coor : Ground position coordinates as (latitude, longitude) in degrees.
    radius : Radius of earths surface at given coordinates.
    ECEFpos : GroundPosition location in ECEF frame.
    alt : Altitude data.
    az : Azimuth data.
    nadirAng : Nadir angle data.
    length : Length of data attributes.
    """

    def __init__(self, name, coor):
        """
        Define instance variables.

        Parameters
        ----------
        name : str
            A tag given to the location for indexing purposes.
        coor : tuple of floats
            Specifies observer position in decimal degrees,
            (latitude, longitude).
        """
        self.name = name
        self.coor = coor
        self.radius = self.getRadius(coor)
        self.ECEFpos = self.getECEF(coor, self.radius)
        self.alt = None
        self.az = None
        self.nadirAng = None
        self.length = None

    def __str__(self):
        """Defines GroundPosition information string."""
        title = 'Celest.GoundPosition Object\n'
        name = f'Name: {self.name}\t'
        coor = f'Coordinates: {self.coor}\t'
        radius = f'Radius: {self.radius}'

        return title + name + coor + radius

    def getRadius(self, obsCoor):
        """
        Instantiates radius attribute.

        Uses the World Geodetic System WGS84 to calculate the radius at the
        given coordinates.

        Parameters
        ----------
        obsCoor : tuple of floats
            Specifies observer position in decimal degrees,
            (latitude, longitude).

        Returns
        -------
        radius : float
            The radius given in km.
        """
        phi = radians(obsCoor[0])

        # Define WGS84 Parameters.
        semiMajor = 6378.137**2
        semiMinor = 6356.752314245**2

        numerator = semiMajor * semiMinor
        denominator = semiMajor * sin(phi)**2 + semiMinor * cos(phi)**2

        return sqrt(numerator / denominator)

    def getECEF(self, obsCoor, radius):
        """
        Instantiates ECEFpos attribute.

        Converts the ground positions geographical coordinates and radius into
        the ECEF cartesian reference frame.

        Parameters
        ----------
        obsCoor : tuple of floats
            Specifies observer position in decimal degrees,
            (latitude, longitude).
        radius : float
            Earth radius at the given coordinates.

        Returns
        -------
        ECEFpos : ndarray
            Array of shape (3,) with X, Y, Z ECEF position data.
        """
        if obsCoor[1] < 0:
            theta = radians(360 + obsCoor[1])
        else:
            theta = radians(obsCoor[1])
        phi = np.radians(90 - obsCoor[0])
        x = radius*cos(theta)*sin(phi)
        y = radius*sin(theta)*sin(phi)
        z = radius*cos(phi)
        ECEFpos = np.array([x, y, z])

        return ECEFpos



"""Satellite orbital coordinate conversions.

Notes
-----
This module contains the Satellite class to calculate different position
representations for Earth orbiting satellites.

Class
-----
Satellite: Object used to collect and calculate orbital representations.
    timeData : Instantiate time attribute with orbital time dependency.
    positionData : Instantiate ECIdata or ECEFdata attribute with orbital
        position data.
    getERA : Instantiate ERAdata attribute.
    getECI : Instantiate ECIdata attribute.
    getECEF : Instantiate ECEFdata attribute.
    getAltAz : Instantiate horizontal attribute.
    getNdrAng : Instantiate nadirAng attribute.
    saveData : Save class data in local directory.
"""


import numpy as np
import pandas as pd
from math import pi, cos, sin, radians
from datetime import datetime, timedelta
import julian


class Satellite:
    """
    Satellite object stores and computes orbital coordinate data. Units are in
    metric.

    Methods
    -------
    timeData : Instantiate time attribute with orbital time dependency.
    positionData : Instantiate ECIdata or ECEFdata attribute with orbital
        position data.
    getERA : Instantiate ERAdata attribute.
    getECI : Instantiate ECIdata attribute.
    getECEF : Instantiate ECEFdata attribute.
    getAltAz : Instantiates the GroundPosition object's .alt and .az
        attributes.
    getNdrAng : Instantiates the GroundPosition object's .nadirAng attribute.
    saveData : Save class data in local directory.

    Instance Variables
    ------------------
    times : Orbital time dependencies.
    ERAdata : Earth rotation angles.
    ECIdata : Earth centered inertial position data.
    ECEFdata : Earth centered earth fixed position data.
    gs : Dictionary with GroundPosition.name as the key and the GroundPosition
        object as the corresponding value.
    length : Length of data attributes.
    """

    def __init__(self):
        """Define instance variables."""

        self.times = None
        self.ERAdata = None
        self.ECIdata = None
        self.ECEFdata = None
        self.gs = {}
        self.length = None

    def timeData(self, timeData):
        """
        Instantiate time attribute with orbital time dependency.

        Takes UTC times.

        Parameters
        ----------
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        None
        """
        self.times = timeData
        self.length = timeData.shape[0]

    def positionData(self, posData, type):
        """
        Instantiate ECIdata or ECEFdata attribute with orbital position data.

        Takes in ECI or ECEF positon data in cartesian coordinates and a data
        type specifier.

        Parameters
        ----------
        posData : ndarray
            Array of shape (n,3) with columns of X, Y, Z position data.
        type : str
            Specifies posData as "ECI" or "ECEF".

        Returns
        -------
        None
        """
        if type == 'ECI':
            self.ECIdata = posData
        elif type == 'ECEF':
            self.ECEFdata = posData

    def getERA(self, **kwargs):
        """
        Instantiate ERAdata attribute.

        Uses times attribute to calculate the Earth Rotation Angle in radians.

        Parameters
        ----------
        None

        **kwargs
        --------
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        self.ERAdata : ndarray
            Array of shape (n,). Contains radian earth rotation angles.

        Usage
        -----
        Must have times attribute initiated or timeData input.
        """
        if type(self.times) == type(None):
            timeData = kwargs['timeData']
            self.timeData(timeData)

        angArr = np.zeros((self.length,))
        Julian = np.zeros((self.length,))

        for i in range(self.length):
            Julian[i] = julian.to_jd(datetime.strptime(self.times[i][:19],
                                                       '%Y-%m-%d %H:%M:%S'))

        # Multiply the elaped time since J2000 by Earth rotation rate and add
        # orientation at J2000.
        dJulian = Julian - 2451545
        angArr = (360.9856123035484 * dJulian + 280.46) % 360

        angArr = np.radians(angArr)
        self.ERAdata = angArr

        return self.ERAdata

    def getECI(self, **kwargs):
        """
        Instantiate ECIdata attribute.

        Computes ECIdata attribute from ECEFdata and ERAdata attribites.

        Parameters
        ----------
        None

        **kwargs
        --------
        posData : ndarray
            Array of shape (n,3) with columns of X, Y, Z position data assumed
            ECEF.
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        self.ECIdata : ndarray
            Array of shape (n,3) with columns of X, Y, Z ECI position data.

        Usage
        -----
        Must have ECEFdata and ERAdata attributes initiated or posData and/or
        timeData inputs.
        """
        if type(self.ERAdata) == type(None):
            self.getERA(**kwargs)
        if type(self.ECEFdata) == type(None):
            ECEFdata = kwargs['posData']
            self.positionData(ECEFdata, 'ECEF')

        # Rotate ECEFdata around z-axis by ERA.
        ECIvec = np.zeros((self.length, 3))
        A11 = np.cos(self.ERAdata)
        A12 = -np.sin(self.ERAdata)
        A21 = np.sin(self.ERAdata)
        A22 = np.cos(self.ERAdata)

        ECIvec[:, 0] = np.add(np.multiply(A11, self.ECEFdata[:, 0]),
                              np.multiply(A12, self.ECEFdata[:, 1]))
        ECIvec[:, 1] = np.add(np.multiply(A21, self.ECEFdata[:, 0]),
                              np.multiply(A22, self.ECEFdata[:, 1]))
        ECIvec[:, 2] = self.ECEFdata[:, 2]

        self.ECIdata = ECIvec

        return self.ECIdata

    def getECEF(self, **kwargs):
        """
        Instantiate ECEFdata attribute.

        Computes ECEFdata attribute from ECIdata and ERAdata attribites.

        Parameters
        ----------
        None

        **kwargs
        --------
        posData : ndarray
            Array of shape (n,3) with columns of X, Y, Z position data assumed
            ECI.
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        self.ECEFdata : ndarray
            Array of shape (n,3) with columns of X, Y, Z ECEF position data.

        Usage
        -----
        Must have ECIdata and ERAdata attributes initiated or posData and/or
        timeData inputs.
        """
        if type(self.ERAdata) == type(None):
            self.getERA(**kwargs)
        if type(self.ECIdata) == type(None):
            ECIdata = kwargs['posData']
            self.positionData(ECIdata, 'ECI')

        # Rotate ECIdata around z-axis by -ERA.
        ECEFvec = np.zeros((self.length, 3))
        A11 = np.cos(-self.ERAdata)
        A12 = -np.sin(-self.ERAdata)
        A21 = np.sin(-self.ERAdata)
        A22 = np.cos(-self.ERAdata)

        ECEFvec[:, 0] = np.add(np.multiply(A11, self.ECIdata[:, 0]),
                               np.multiply(A12, self.ECIdata[:, 1]))
        ECEFvec[:, 1] = np.add(np.multiply(A21, self.ECIdata[:, 0]),
                               np.multiply(A22, self.ECIdata[:, 1]))
        ECEFvec[:, 2] = self.ECIdata[:, 2]

        self.ECEFdata = ECEFvec

        return self.ECEFdata

    def getAltAz(self, groundPos, **kwargs):
        """
        Instantiate horizontal attribute.

        This method takes in a GroundPosition object and instantiates its alt
        and az attributes using the satellites position data. The
        GroundPosition object is then stored as a value in the gs attribute's
        dictionary with the corresponding key being its name attribute.

        Parameters
        ----------
        groundPos : GroundPosition object
            GroundPosition object instantiated as per its documentation.

        **kwargs
        --------
        posData : ndarray
            Array of shape (n,3) with columns of X, Y, Z position data assumed
            ECEF.
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        AltAz : tuple of ndarrays
            (alt, az) tuple containing the computed altitude and azimuth data.
            Both alt and az are both arrays of shape (n,).

        Usage
        -----
        Must have ECEFdata attribute initiated or the posData and timeData
        inputs.
        """
        def getAngle(vecOne, vecTwo):
            """
            Calculate degree angle bewteen two vectors.

            Takes two multidimensional cartesian coordinate vectors and returns
            degree angle between the two arrays element-wise.

            Parameters
            ----------
            vecOne : ndarray
                Array of shape (n,3) representing n XYZ vectors.
            vecTwo : ndarray
                Array of shape (n,3) representing n XYZ vectors.

            Returns
            -------
            ang : float
                Degree angle between vecOne and vecTwo.
            """
            # Use simple linalg formula.
            dividend = np.einsum('ij, ij->i', vecOne, vecTwo)
            divisor = np.multiply(np.linalg.norm(vecOne, axis=1),
                                  np.linalg.norm(vecTwo, axis=1))
            arg = np.divide(dividend, divisor)
            ang = np.degrees(np.arccos(arg))

            return ang

        if type(self.ECEFdata) == type(None):
            self.getECEF(**kwargs)

        if groundPos.name not in self.gs:
            groundPos.length = self.length
            self.gs[groundPos.name] = groundPos

        # Convert observer position into spherical then cartesian.
        obsCoor = groundPos.coor
        radius = groundPos.radius
        if obsCoor[1] < 0:
            theta = radians(360 + obsCoor[1])
        else:
            theta = radians(obsCoor[1])
        phi = np.radians(90 - obsCoor[0])
        x = radius*cos(theta)*sin(phi)
        y = radius*sin(theta)*sin(phi)
        z = radius*cos(phi)
        xyzObs = np.full((self.length, 3), np.array([x, y, z]))

        # Determine line of sight vector then altitude.
        ECEFvec = self.ECEFdata
        xyzLOS = np.subtract(ECEFvec, xyzObs)
        self.gs[groundPos.name].alt = 90 - getAngle(xyzLOS, xyzObs)

        # Find surface tangent vector passing through z-axis.
        kHat = np.full((self.length, 3), np.array([0, 0, 1]))
        beta = pi/2 - phi
        tangentVec = np.subtract((kHat.T * radius/np.sin(beta)).T, xyzObs)

        # Find LOS projection on tangent plane.
        coeff = np.einsum('ij, ij->i', xyzLOS, xyzObs)/radius**2
        normProj = (xyzObs.T * coeff).T
        projLOS = np.subtract(xyzLOS, normProj)

        # Determing aECIzimuth.
        vecOne = np.cross(tangentVec, xyzObs)
        normOne = 1/np.linalg.norm(vecOne, axis=1).reshape((self.length, 1))
        vecOneUnit = normOne*vecOne
        vecTwo = (vecOneUnit.T * np.einsum('ij, ij->i', projLOS, vecOneUnit)).T
        normTwo = 1/np.linalg.norm(vecTwo, axis=1).reshape((self.length, 1))
        vecTwoUnit = normTwo*vecTwo

        posEq = np.isclose(vecOneUnit, vecTwoUnit, atol=0.01)
        negEq = np.isclose(vecOneUnit, -vecTwoUnit, atol=0.01)
        posInd = np.where(np.all(posEq, axis=1))[0]
        negInd = np.where(np.all(negEq, axis=1))[0]

        Az = np.zeros((self.length,))
        Az[posInd] = getAngle(tangentVec[posInd], projLOS[posInd])
        Az[negInd] = 360 - getAngle(tangentVec[negInd], projLOS[negInd])

        self.gs[groundPos.name].az = Az

        return self.gs[groundPos.name].alt, self.gs[groundPos.name].az

    def getNdrAng(self, groundPos, **kwargs):
        """
        Return Nadir-LOS angle array.

        This method takes in a GroundPosition object and instantiates its
        nadirAng attribute with the angle made between the satellites nadir and
        line-of-sight (from the observer to the ground station) vectors. The
        GroundPosition object is then stored as a value in the gs attribute's
        dictionary with the corresponding key being its name attribute.

        Parameters
        ----------
        groundPos : GroundPosition object
            GroundPosition object instantiated as per its documentation.

        **kwargs
        --------
        posData : ndarray
            Array of shape (n,3) with columns of X, Y, Z position data assumed
            ECI.
        timeData : ndarray
            Array of shape (n,). Contains datetime objects in UTC.

        Returns
        -------
        nadirAng : ndarray
            Array of shape (n,) with nadir-LOS angle data.

        Usage
        -----
        Must have ECEFdata attribute initiated or the posData and timeData
        inputs.
        """
        if type(self.ECEFdata) == type(None):
            self.getECEF(**kwargs)

        if groundPos.name not in self.gs:
            groundPos.length = self.length
            self.gs[groundPos.name] = groundPos

        obsCoor = groundPos.coor
        radius = groundPos.radius
        if obsCoor[1] < 0:
            theta = radians(360 + obsCoor[1])
        else:
            theta = radians(obsCoor[1])
        phi = np.radians(90 - obsCoor[0])
        x = radius*cos(theta)*sin(phi)
        y = radius*sin(theta)*sin(phi)
        z = radius*cos(phi)
        xyzObs = np.full((self.length, 3), np.array([x, y, z]))

        ECEFvec = self.ECEFdata
        xyzLOS = np.subtract(ECEFvec, xyzObs)

        # Use simple linalg formula.
        dividend = np.einsum('ij, ij->i', xyzLOS, ECEFvec)
        divisor = np.multiply(np.linalg.norm(xyzLOS, axis=1),
                              np.linalg.norm(ECEFvec, axis=1))
        arg = np.divide(dividend, divisor)
        ang = np.degrees(np.arccos(arg))

        self.gs[groundPos.name].nadirAng = ang

        return self.gs[groundPos.name].nadirAng

    def saveData(self, fileName, delimiter):
        """
        Save satellite data to local directory.

        Parameters
        ----------
        fileName : str
            File name of output file. Can be a .txt or .csv file.
        delimiter : str
            String of length 1 representing the feild delimiter for the output
            file.

        Returns
        -------
        None

        Usage
        -----
        It is recommended to use a tab delimiter for .txt files and comma
        delimiters for .csv files. Will return an error if fileName already
        exists in the current working directory.
        """
        data = {}

        if type(self.times) != type(None):
            data['Time (UTC)'] = pd.Series(self.times)

        if type(self.ERAdata) != type(None):
            data['ERA (rad)'] = pd.Series(self.ERAdata)

        if type(self.ECIdata) != type(None):
            data['ECI.X'] = pd.Series(self.ECIdata[:, 0])
            data['ECI.Y'] = pd.Series(self.ECIdata[:, 1])
            data['ECI.Z'] = pd.Series(self.ECIdata[:, 2])

        if type(self.ECEFdata) != type(None):
            data['ECEF.X'] = pd.Series(self.ECEFdata[:, 0])
            data['ECEF.Y'] = pd.Series(self.ECEFdata[:, 1])
            data['ECEF.Z'] = pd.Series(self.ECEFdata[:, 2])

        for key in self.gs:
            if type(self.gs[key].alt) != type(None):
                data[f'{key}.alt'] = self.gs[key].alt
                data[f'{key}.az'] = self.gs[key].az
            if type(self.gs[key].nadirAng) != type(None):
                data[f'{key}.NdrAng'] = self.gs[key].nadirAng

        df = pd.DataFrame(data)
        df.to_csv(fileName, sep=delimiter)
